@@ -18,6 +18,7 @@ from typing import (
 import aiohttp
 from aiodocker.docker import Docker, DockerContainer
 from aiodocker.exceptions import DockerError
+import async_timeout
 import psutil
 
 from ai.backend.common.logging import BraceStyleAdapter
@@ -34,7 +35,7 @@ from .resources import (
 )
 from .. import __version__
 from ..resources import (
-    AbstractAllocMap,
+    AbstractAllocMap, DeviceSlotInfo,
     DiscretePropertyAllocMap,
     AbstractComputeDevice,
     AbstractComputePlugin,
@@ -50,8 +51,8 @@ log = BraceStyleAdapter(logging.getLogger(__name__))
 
 
 async def fetch_api_stats(container: DockerContainer) -> Optional[Dict[str, Any]]:
+    short_cid = container._id[:7]
     try:
-        short_cid = container._id[:7]
         ret = await container.stats(stream=False)  # TODO: cache
     except RuntimeError as e:
         msg = str(e.args[0]).lower()
@@ -116,7 +117,7 @@ class CPUPlugin(AbstractComputePlugin):
     async def list_devices(self) -> Collection[CPUDevice]:
         cores = await libnuma.get_available_cores()
         overcommit_factor = int(os.environ.get('BACKEND_CPU_OVERCOMMIT_FACTOR', '1'))
-        assert 1 <= overcommit_factor <= 4
+        assert 1 <= overcommit_factor <= 10
         return [
             CPUDevice(
                 device_id=DeviceId(str(core_idx)),
@@ -188,7 +189,11 @@ class CPUPlugin(AbstractComputePlugin):
 
         async def api_impl(container_id):
             container = DockerContainer(ctx.agent.docker, id=container_id)
-            ret = await fetch_api_stats(container)
+            try:
+                async with async_timeout.timeout(2.0):
+                    ret = await fetch_api_stats(container)
+            except asyncio.TimeoutError:
+                return None
             if ret is None:
                 return None
             cpu_used = nmget(ret, 'cpu_stats.cpu_usage.total_usage', 0) / 1e6
@@ -198,6 +203,8 @@ class CPUPlugin(AbstractComputePlugin):
             impl = sysfs_impl
         elif ctx.mode == StatModes.DOCKER:
             impl = api_impl
+        else:
+            raise RuntimeError("should not reach here")
 
         q = Decimal('0.000')
         per_container_cpu_used = {}
@@ -229,8 +236,12 @@ class CPUPlugin(AbstractComputePlugin):
     async def create_alloc_map(self) -> AbstractAllocMap:
         devices = await self.list_devices()
         return DiscretePropertyAllocMap(
-            devices=devices,
-            prop_func=lambda dev: dev.processing_units)
+            device_slots={
+                dev.device_id:
+                    DeviceSlotInfo(SlotTypes.COUNT, SlotName('cpu'), Decimal(dev.processing_units))
+                for dev in devices
+            },
+        )
 
     async def get_hooks(self, distro: str, arch: str) -> Sequence[Path]:
         # TODO: move the sysconf hook in libbaihook.so here
@@ -452,7 +463,11 @@ class MemoryPlugin(AbstractComputePlugin):
 
         async def api_impl(container_id):
             container = DockerContainer(ctx.agent.docker, id=container_id)
-            ret = await fetch_api_stats(container)
+            try:
+                async with async_timeout.timeout(2.0):
+                    ret = await fetch_api_stats(container)
+            except asyncio.TimeoutError:
+                return None
             if ret is None:
                 return None
             mem_cur_bytes = nmget(ret, 'memory_stats.usage', 0)
@@ -472,6 +487,8 @@ class MemoryPlugin(AbstractComputePlugin):
             impl = sysfs_impl
         elif ctx.mode == StatModes.DOCKER:
             impl = api_impl
+        else:
+            raise RuntimeError("should not reach here")
 
         per_container_mem_used_bytes = {}
         per_container_io_read_bytes = {}
@@ -526,8 +543,12 @@ class MemoryPlugin(AbstractComputePlugin):
     async def create_alloc_map(self) -> AbstractAllocMap:
         devices = await self.list_devices()
         return DiscretePropertyAllocMap(
-            devices=devices,
-            prop_func=lambda dev: dev.memory_size)
+            device_slots={
+                dev.device_id:
+                    DeviceSlotInfo(SlotTypes.BYTES, SlotName('mem'), Decimal(dev.memory_size))
+                for dev in devices
+            },
+        )
 
     async def get_hooks(self, distro: str, arch: str) -> Sequence[Path]:
         return []
