@@ -5,26 +5,28 @@ import functools
 import importlib
 from ipaddress import ip_network, _BaseAddress as BaseIPAddress
 import logging, logging.config
-import multiprocessing
 import os, os.path
 from pathlib import Path
 from pprint import pformat, pprint
 import signal
 import sys
-import time
-import traceback
 from typing import (
     Any,
+    AsyncGenerator,
     Callable,
     ClassVar,
+    Coroutine,
     Dict,
+    Literal,
     Mapping,
+    Sequence,
     Set,
-    TYPE_CHECKING,
+    Tuple,
     cast,
 )
 from uuid import UUID
 
+import aiomonitor
 import aiotools
 from aiotools import aclosing
 from callosum.rpc import Peer, RPCMessage
@@ -39,8 +41,12 @@ from ai.backend.common.etcd import AsyncEtcd, ConfigScopes
 from ai.backend.common.logging import Logger, BraceStyleAdapter
 from ai.backend.common.plugin.monitor import ErrorPluginContext, StatsPluginContext
 from ai.backend.common.types import (
-    aobject, HostPortPair, KernelId,
+    HardwareMetadata, aobject,
+    ClusterInfo,
+    HostPortPair,
+    KernelId,
     KernelCreationConfig,
+    SessionId,
 )
 from ai.backend.common.utils import current_loop
 from . import __version__ as VERSION
@@ -49,16 +55,15 @@ from .config import (
     agent_local_config_iv,
     agent_etcd_config_iv,
     k8s_extra_config_iv,
+    docker_extra_config_iv,
     registry_local_config_iv,
     registry_ecr_config_iv,
     container_etcd_config_iv,
 )
+from .exception import ResourceError
 from .types import AgentBackend, VolumeInfo, LifecycleEvent
 from .utils import get_subnet_ip
 
-if TYPE_CHECKING:
-    from typing import Any, Dict, Optional  # noqa
-
 log = BraceStyleAdapter(logging.getLogger('ai.backend.agent.server'))
 
 deeplearning_image_keys = {
@@ -71,6 +76,8 @@ deeplearning_sample_volume = VolumeInfo(
     'deeplearning-samples', '/home/work/samples', 'ro',
 )
 
+agent_instance: AgentRPCServer
+
 
 async def get_extra_volumes(docker, lang):
     avail_volumes = (await docker.volumes.list())['Volumes']
@@ -98,29 +105,13 @@ async def get_extra_volumes(docker, lang):
     return mount_list
 
 
-def update_last_used(meth: Callable) -> Callable:
-    @functools.wraps(meth)
-    async def _inner(self, raw_kernel_id: str, *args, **kwargs):
-        try:
-            kernel_obj = self.agent.kernel_registry[KernelId(UUID(raw_kernel_id))]
-            kernel_obj.last_used = time.monotonic()
-        except KeyError:
-            pass
-        return await meth(self, raw_kernel_id, *args, **kwargs)
-    return _inner
-
-
 def collect_error(meth: Callable) -> Callable:
     @functools.wraps(meth)
-    async def _inner(self, *args, **kwargs):
+    async def _inner(self: AgentRPCServer, *args, **kwargs):
         try:
             return await meth(self, *args, **kwargs)
         except Exception:
-            exc_type, exc, tb = sys.exc_info()
-            pretty_message = ''.join(traceback.format_exception_only(exc_type, exc)).strip()
-            pretty_tb = ''.join(traceback.format_tb(tb)).strip()
-            # TODO: Determine severity
-            await self.agent.produce_event('agent_error', pretty_message, pretty_tb)
+            await self.agent.produce_error_event()
             raise
     return _inner
 
@@ -132,24 +123,30 @@ class RPCFunctionRegistry:
     def __init__(self) -> None:
         self.functions = set()
 
-    def __call__(self, meth: Callable) -> Callable[[RPCMessage], Any]:
+    def __call__(
+        self,
+        meth: Callable[..., Coroutine[None, None, Any]],
+    ) -> Callable[[AgentRPCServer, RPCMessage], Coroutine[None, None, Any]]:
 
         @functools.wraps(meth)
-        async def _inner(self: AgentRPCServer, request: RPCMessage):
+        async def _inner(self_: AgentRPCServer, request: RPCMessage) -> Any:
             try:
                 if request.body is None:
-                    return await meth(self)
+                    return await meth(self_)
                 else:
                     return await meth(
-                        self,
+                        self_,
                         *request.body['args'],
                         **request.body['kwargs'],
                     )
             except (asyncio.CancelledError, asyncio.TimeoutError):
                 raise
+            except ResourceError:
+                # This is an expected scenario.
+                raise
             except Exception:
                 log.exception('unexpected error')
-                await self.error_monitor.capture_exception()
+                await self_.error_monitor.capture_exception()
                 raise
 
         self.functions.add(meth.__name__)
@@ -182,6 +179,8 @@ class AgentRPCServer(aobject):
 
     async def __ainit__(self) -> None:
         # Start serving requests.
+        self._create_sema = asyncio.Semaphore(4)
+        self._destroy_sema = asyncio.Semaphore(4)
         await self.update_status('starting')
 
         if not self.skip_detect_manager:
@@ -246,8 +245,11 @@ class AgentRPCServer(aobject):
         self.local_config['vfolder'] = config.vfolder_config_iv.check(
             await self.etcd.get_prefix('volumes')
         )
-        log.info('configured vfolder mount base: {0}', self.local_config['vfolder']['mount'])
-        log.info('configured vfolder fs prefix: {0}', self.local_config['vfolder']['fsprefix'])
+        if self.local_config['vfolder']['mount'] is None:
+            log.info('assuming use of storage-proxy since vfolder mount path is not configured in etcd')
+        else:
+            log.info('configured vfolder mount base: {0}', self.local_config['vfolder']['mount'])
+            log.info('configured vfolder fs prefix: {0}', self.local_config['vfolder']['fsprefix'])
 
         # Fill up shared agent configurations from etcd.
         agent_etcd_config = agent_etcd_config_iv.check(
@@ -289,59 +291,99 @@ class AgentRPCServer(aobject):
     @rpc_function
     @collect_error
     async def ping(self, msg: str) -> str:
+        log.debug('rpc::ping()')
         return msg
 
     @rpc_function
-    @update_last_used
+    @collect_error
+    async def gather_hwinfo(self) -> Mapping[str, HardwareMetadata]:
+        log.debug('rpc::gather_hwinfo()')
+        return await self.agent.gather_hwinfo()
+
+    @rpc_function
     @collect_error
     async def ping_kernel(self, kernel_id: str):
         log.debug('rpc::ping_kernel({0})', kernel_id)
 
     @rpc_function
-    @update_last_used
     @collect_error
-    async def create_kernel(self, kernel_id: str, config: dict):
-        log.info('rpc::create_kernel(k:{0}, img:{1})',
-                 kernel_id, config['image']['canonical'])
-        result = await self.agent.create_kernel(
-            KernelId(UUID(kernel_id)), cast(KernelCreationConfig, config))
-        return {
-            'id': str(result['id']),
-            'kernel_host': result['kernel_host'],
-            'repl_in_port': result['repl_in_port'],
-            'repl_out_port': result['repl_out_port'],
-            'stdin_port': result['stdin_port'],    # legacy
-            'stdout_port': result['stdout_port'],  # legacy
-            'service_ports': result['service_ports'],
-            'container_id': result['container_id'],
-            'resource_spec': result['resource_spec'],
-            'attached_devices': result['attached_devices'],
-        }
+    async def create_kernels(
+        self,
+        creation_id: str,
+        raw_session_id: str,
+        raw_kernel_ids: Sequence[str],
+        raw_configs: Sequence[dict],
+        raw_cluster_info: dict,
+    ):
+        async with self._create_sema:
+            cluster_info = cast(ClusterInfo, raw_cluster_info)
+            session_id = SessionId(UUID(raw_session_id))
+            raw_results = []
+            coros = []
+            for raw_kernel_id, raw_config in zip(raw_kernel_ids, raw_configs):
+                log.info('rpc::create_kernel(k:{0}, img:{1})',
+                        raw_kernel_id, raw_config['image']['canonical'])
+                kernel_id = KernelId(UUID(raw_kernel_id))
+                kernel_config = cast(KernelCreationConfig, raw_config)
+                coros.append(self.agent.create_kernel(
+                    creation_id,
+                    session_id,
+                    kernel_id,
+                    kernel_config,
+                    cluster_info,
+                ))
+            results = await asyncio.gather(*coros, return_exceptions=True)
+        errors = [*filter(lambda item: isinstance(item, Exception), results)]
+        if errors:
+            # Raise up the first error.
+            if len(errors) == 1:
+                raise errors[0]
+            raise aiotools.MultiError("agent.create_kernels() failed", errors=errors)
+        raw_results = [
+            {
+                'id': str(result['id']),
+                'kernel_host': result['kernel_host'],
+                'repl_in_port': result['repl_in_port'],
+                'repl_out_port': result['repl_out_port'],
+                'stdin_port': result['stdin_port'],    # legacy
+                'stdout_port': result['stdout_port'],  # legacy
+                'service_ports': result['service_ports'],
+                'container_id': result['container_id'],
+                'resource_spec': result['resource_spec'],
+                'attached_devices': result['attached_devices'],
+            }
+            for result in results
+        ]
+        return raw_results
 
     @rpc_function
-    @update_last_used
     @collect_error
-    async def destroy_kernel(self, kernel_id: str, reason: str = None):
-        log.info('rpc::destroy_kernel(k:{0})', kernel_id)
-        done = asyncio.Event()
-        await self.agent.inject_container_lifecycle_event(
-            KernelId(UUID(kernel_id)),
-            LifecycleEvent.DESTROY,
-            reason or 'user-requested',
-            done_event=done,
-        )
-        await done.wait()
+    async def destroy_kernel(
+        self,
+        kernel_id: str,
+        reason: str = None,
+        suppress_events: bool = False,
+    ):
+        async with self._destroy_sema:
+            log.info('rpc::destroy_kernel(k:{0})', kernel_id)
+            done = asyncio.Event()
+            await self.agent.inject_container_lifecycle_event(
+                KernelId(UUID(kernel_id)),
+                LifecycleEvent.DESTROY,
+                reason or 'user-requested',
+                done_event=done,
+                suppress_events=suppress_events,
+            )
+            await done.wait()
         return getattr(done, '_result', None)
 
     @rpc_function
-    @update_last_used
     @collect_error
     async def interrupt_kernel(self, kernel_id: str):
         log.info('rpc::interrupt_kernel(k:{0})', kernel_id)
         await self.agent.interrupt_kernel(KernelId(UUID(kernel_id)))
 
     @rpc_function
-    @update_last_used
     @collect_error
     async def get_completions(self, kernel_id: str,
                               text: str, opts: dict):
@@ -349,32 +391,40 @@ class AgentRPCServer(aobject):
         await self.agent.get_completions(KernelId(UUID(kernel_id)), text, opts)
 
     @rpc_function
-    @update_last_used
     @collect_error
     async def get_logs(self, kernel_id: str):
         log.info('rpc::get_logs(k:{0})', kernel_id)
         return await self.agent.get_logs(KernelId(UUID(kernel_id)))
 
     @rpc_function
-    @update_last_used
     @collect_error
-    async def restart_kernel(self, kernel_id: str, new_config: dict):
-        log.info('rpc::restart_kernel(k:{0})', kernel_id)
+    async def restart_kernel(
+        self,
+        creation_id: str,
+        session_id: str,
+        kernel_id: str,
+        updated_config: dict,
+    ):
+        log.info('rpc::restart_kernel(s:{0}, k:{1})', session_id, kernel_id)
         return await self.agent.restart_kernel(
-            KernelId(UUID(kernel_id)), cast(KernelCreationConfig, new_config))
+            creation_id,
+            SessionId(UUID(session_id)),
+            KernelId(UUID(kernel_id)),
+            cast(KernelCreationConfig, updated_config),
+        )
 
     @rpc_function
-    @update_last_used
     @collect_error
-    async def execute(self,
-                      kernel_id,          # type: str
-                      api_version,        # type: int
-                      run_id,             # type: str
-                      mode,               # type: str
-                      code,               # type: str
-                      opts,               # type: Dict[str, Any]
-                      flush_timeout,      # type: float
-                      ):
+    async def execute(
+        self,
+        kernel_id,          # type: str
+        api_version,        # type: int
+        run_id,             # type: str
+        mode,               # type: Literal['query', 'batch', 'continue', 'input']
+        code,               # type: str
+        opts,               # type: Dict[str, Any]
+        flush_timeout,      # type: float
+    ):
         # type: (...) -> Dict[str, Any]
         if mode != 'continue':
             log.info('rpc::execute(k:{0}, run-id:{1}, mode:{2}, code:{3!r})',
@@ -382,7 +432,9 @@ class AgentRPCServer(aobject):
                      code[:20] + '...' if len(code) > 20 else code)
         result = await self.agent.execute(
             KernelId(UUID(kernel_id)),
-            run_id, mode, code,
+            run_id,
+            mode,
+            code,
             opts=opts,
             api_version=api_version,
             flush_timeout=flush_timeout
@@ -390,46 +442,59 @@ class AgentRPCServer(aobject):
         return result
 
     @rpc_function
-    @update_last_used
     @collect_error
-    async def start_service(self,
-                            kernel_id,   # type: str
-                            service,     # type: str
-                            opts         # type: Dict[str, Any]
-                            ):
+    async def execute_batch(
+        self,
+        kernel_id,          # type: str
+        startup_command,    # type: str
+    ) -> None:
+        # DEPRECATED
+        asyncio.create_task(self.agent.execute_batch(
+            KernelId(UUID(kernel_id)),
+            startup_command,
+        ))
+        await asyncio.sleep(0)
+
+    @rpc_function
+    @collect_error
+    async def start_service(
+        self,
+        kernel_id,   # type: str
+        service,     # type: str
+        opts         # type: Dict[str, Any]
+    ):
         # type: (...) -> Dict[str, Any]
         log.info('rpc::start_service(k:{0}, app:{1})', kernel_id, service)
         return await self.agent.start_service(KernelId(UUID(kernel_id)), service, opts)
 
     @rpc_function
-    @update_last_used
+    @collect_error
+    async def shutdown_service(
+        self,
+        kernel_id,  # type: str
+        service,    # type: str
+    ):
+        log.info('rpc::shutdown_service(k:{0}, app:{1})', kernel_id, service)
+        return await self.agent.shutdown_service(KernelId(UUID(kernel_id)), service)
+
+    @rpc_function
     @collect_error
     async def upload_file(self, kernel_id: str, filename: str, filedata: bytes):
         log.info('rpc::upload_file(k:{0}, fn:{1})', kernel_id, filename)
         await self.agent.accept_file(KernelId(UUID(kernel_id)), filename, filedata)
 
     @rpc_function
-    @update_last_used
     @collect_error
     async def download_file(self, kernel_id: str, filepath: str):
         log.info('rpc::download_file(k:{0}, fn:{1})', kernel_id, filepath)
         return await self.agent.download_file(KernelId(UUID(kernel_id)), filepath)
 
     @rpc_function
-    @update_last_used
     @collect_error
     async def list_files(self, kernel_id: str, path: str):
         log.info('rpc::list_files(k:{0}, fn:{1})', kernel_id, path)
         return await self.agent.list_files(KernelId(UUID(kernel_id)), path)
 
-    @rpc_function
-    @update_last_used
-    @collect_error
-    async def refresh_idle(self, kernel_id: str):
-        # update_last_used decorator already implements this. :)
-        log.debug('rpc::refresh_idle(k:{})', kernel_id)
-        pass
-
     @rpc_function
     @collect_error
     async def shutdown_agent(self, terminate_kernels: bool):
@@ -437,6 +502,30 @@ class AgentRPCServer(aobject):
         log.info('rpc::shutdown_agent()')
         pass
 
+    @rpc_function
+    @collect_error
+    async def create_overlay_network(self, network_name: str) -> None:
+        log.debug('rpc::create_overlay_network(name:{})', network_name)
+        return await self.agent.create_overlay_network(network_name)
+
+    @rpc_function
+    @collect_error
+    async def destroy_overlay_network(self, network_name: str) -> None:
+        log.debug('rpc::destroy_overlay_network(name:{})', network_name)
+        return await self.agent.destroy_overlay_network(network_name)
+
+    @rpc_function
+    @collect_error
+    async def create_local_network(self, network_name: str) -> None:
+        log.debug('rpc::create_local_network(name:{})', network_name)
+        return await self.agent.create_local_network(network_name)
+
+    @rpc_function
+    @collect_error
+    async def destroy_local_network(self, network_name: str) -> None:
+        log.debug('rpc::destroy_local_network(name:{})', network_name)
+        return await self.agent.destroy_local_network(network_name)
+
     @rpc_function
     @collect_error
     async def reset_agent(self):
@@ -455,7 +544,11 @@ class AgentRPCServer(aobject):
 
 
 @aiotools.server
-async def server_main_logwrapper(loop, pidx, _args):
+async def server_main_logwrapper(
+    loop: asyncio.AbstractEventLoop,
+    pidx: int,
+    _args: Tuple[Any, ...],
+) -> AsyncGenerator[None, signal.Signals]:
     setproctitle(f"backend.ai: agent worker-{pidx}")
     log_endpoint = _args[1]
     logger = Logger(_args[0]['logging'], is_master=False, log_endpoint=log_endpoint)
@@ -465,14 +558,18 @@ async def server_main_logwrapper(loop, pidx, _args):
 
 
 @aiotools.server
-async def server_main(loop, pidx, _args):
+async def server_main(
+    loop: asyncio.AbstractEventLoop,
+    pidx: int,
+    _args: Tuple[Any, ...],
+) -> AsyncGenerator[None, signal.Signals]:
     local_config = _args[0]
 
     log.info('Preparing kernel runner environments...')
     kernel_mod = importlib.import_module(
         f"ai.backend.agent.{local_config['agent']['backend'].value}.kernel",
     )
-    krunner_volumes = await kernel_mod.prepare_krunner_env(local_config)
+    krunner_volumes = await kernel_mod.prepare_krunner_env(local_config)  # type: ignore
     # TODO: merge k8s branch: nfs_mount_path = local_config['baistatic']['mounted-at']
     log.info('Kernel runner environments: {}', [*krunner_volumes.keys()])
     local_config['container']['krunner-volumes'] = krunner_volumes
@@ -525,6 +622,16 @@ async def server_main(loop, pidx, _args):
     # Pre-load compute plugin configurations.
     local_config['plugins'] = await etcd.get_prefix_dict('config/plugins/accelerator')
 
+    # Start aiomonitor.
+    # Port is set by config (default=50002).
+    monitor = aiomonitor.Monitor(
+        loop,
+        port=local_config['agent']['aiomonitor-port'],
+        console_enabled=False
+    )
+    monitor.prompt = "monitor (agent) >>> "
+    monitor.start()
+
     # Start RPC server.
     global agent_instance
     agent = await AgentRPCServer.new(
@@ -534,9 +641,12 @@ async def server_main(loop, pidx, _args):
     agent_instance = agent
 
     # Run!
-    async with agent:
-        stop_signal = yield
-        agent.mark_stop_signal(stop_signal)
+    try:
+        async with agent:
+            stop_signal = yield
+            agent.mark_stop_signal(stop_signal)
+    finally:
+        monitor.close()
 
 
 @click.group(invoke_without_command=True)
@@ -593,7 +703,8 @@ def main(
 
             registry_cfg = config.check(cfg['registry'], registry_target_config_iv)
             cfg['registry'] = registry_cfg
-
+        if cfg['agent']['backend'] == AgentBackend.DOCKER:
+            config.check(raw_cfg, docker_extra_config_iv)
         if 'debug' in cfg and cfg['debug']['enabled']:
             print('== Agent configuration ==')
             pprint(cfg)
@@ -616,8 +727,6 @@ def main(
               file=sys.stderr)
         raise click.Abort()
 
-    multiprocessing.set_start_method('spawn')
-
     if cli_ctx.invoked_subcommand is None:
 
         if cfg['debug']['coredump']['enabled']:
@@ -656,10 +765,11 @@ def main(
                     import uvloop
                     uvloop.install()
                     log.info('Using uvloop as the event loop backend')
-                aiotools.start_server(server_main_logwrapper,
-                                      num_workers=1,
-                                      use_threading=False,
-                                      args=(cfg, log_endpoint))
+                aiotools.start_server(
+                    server_main_logwrapper,
+                    num_workers=1,
+                    args=(cfg, log_endpoint),
+                )
                 log.info('exit.')
         finally:
             if cfg['agent']['pid-file'].is_file():
