@@ -133,6 +133,7 @@ async def handle_fstab_detail(request: web.Request) -> web.Response:
     log.info('HANDLE_FSTAB_DETAIL')
     params = request.query
     fstab_path = params.get('fstab_path', '/etc/fstab')
+    # FIXME: Remove ignore if https://github.com/python/typeshed/pull/4650 is released
     async with aiofiles.open(fstab_path, mode='r') as fp:  # type: ignore
         content = await fp.read()
         return web.Response(text=content)
@@ -180,6 +181,7 @@ async def handle_mount(request: web.Request) -> web.Response:
     log.info('Mounted ' + params['name'] + ' on ' + mount_prefix)
     if params['edit_fstab']:
         fstab_path = params['fstab_path'] if params['fstab_path'] else '/etc/fstab'
+        # FIXME: Remove ignore if https://github.com/python/typeshed/pull/4650 is released
         async with aiofiles.open(fstab_path, mode='r+') as fp:  # type: ignore
             fstab = Fstab(fp)
             await fstab.add(params['fs_location'], str(mountpoint),
@@ -213,6 +215,7 @@ async def handle_umount(request: web.Request) -> web.Response:
         pass
     if params['edit_fstab']:
         fstab_path = params['fstab_path'] if params['fstab_path'] else '/etc/fstab'
+        # FIXME: Remove ignore if https://github.com/python/typeshed/pull/4650 is released
         async with aiofiles.open(fstab_path, mode='r+') as fp:  # type: ignore
             fstab = Fstab(fp)
             await fstab.remove_by_mountpoint(str(mountpoint))
@@ -372,8 +375,9 @@ def main(cli_ctx, config_path, debug):
         log_config.debug('debug mode enabled.')
 
         aiotools.start_server(
-            watcher_server, num_workers=1,
-            use_threading=True, args=(cfg, ),
+            watcher_server,
+            num_workers=1,
+            args=(cfg, ),
             stop_signals={signal.SIGINT, signal.SIGTERM, signal.SIGALRM},
         )
         log.info('exit.')
