@@ -11,13 +11,19 @@ import signal
 import sys
 import time
 from typing import (
-    ClassVar,
-    Optional, Union,
     Awaitable,
-    Sequence, List,
-    MutableMapping, Dict,
+    ClassVar,
+    Dict,
+    List,
+    Mapping,
+    MutableMapping,
+    Optional,
+    Sequence,
+    Union,
 )
+import uuid
 
+from async_timeout import timeout
 import janus
 from jupyter_client import KernelManager
 from jupyter_client.kernelspec import KernelSpecManager
@@ -30,7 +36,8 @@ from .logging import BraceStyleAdapter, setup_logger
 from .utils import wait_local_port_open
 from .compat import current_loop
 from .intrinsic import (
-    init_sshd_service, prepare_sshd_service,
+    init_sshd_service,
+    prepare_sshd_service,
     prepare_ttyd_service,
     prepare_vscode_service,
 )
@@ -60,11 +67,11 @@ async def pipe_output(stream, outsock, target, log_fd):
         log.exception('unexpected error')
 
 
-async def terminate_and_wait(proc: asyncio.subprocess.Process) -> None:
+async def terminate_and_wait(proc: asyncio.subprocess.Process, timeout: float = 2.0) -> None:
     try:
         proc.terminate()
         try:
-            await asyncio.wait_for(proc.wait(), timeout=2.0)
+            await asyncio.wait_for(proc.wait(), timeout=timeout)
         except asyncio.TimeoutError:
             proc.kill()
             await proc.wait()
@@ -72,6 +79,18 @@ async def terminate_and_wait(proc: asyncio.subprocess.Process) -> None:
         pass
 
 
+def promote_path(path_env: str, path_to_promote: Union[Path, str]) -> str:
+    paths = path_env.split(':')
+    print(f"promote_path: {path_to_promote=} {path_env=}", file=sys.stderr)
+    path_to_promote = str(path_to_promote)
+    result_paths = [
+        p for p in paths
+        if path_to_promote != p
+    ]
+    result_paths.insert(0, path_to_promote)
+    return ":".join(result_paths)
+
+
 class BaseRunner(metaclass=ABCMeta):
 
     log_prefix: ClassVar[str] = 'generic-kernel'
@@ -105,8 +124,11 @@ class BaseRunner(metaclass=ABCMeta):
         self.subproc = None
         self.runtime_path = runtime_path
 
+        default_child_env_path = self.default_child_env.pop("PATH", None)
         self.child_env = {**os.environ, **self.default_child_env}
-        self._service_lock = asyncio.Lock()
+        if default_child_env_path is not None and "PATH" not in self.child_env:
+            # set the default PATH env-var only when it's missing from the image
+            self.child_env["PATH"] = default_child_env_path
         config_dir = Path('/home/config')
         try:
             evdata = (config_dir / 'environ.txt').read_text()
@@ -119,6 +141,10 @@ class BaseRunner(metaclass=ABCMeta):
         except Exception:
             log.exception('Reading /home/config/environ.txt failed!')
 
+        # Add ~/.local/bin to the default PATH
+        self.child_env["PATH"] += os.pathsep + '~/.local/bin'
+        os.environ["PATH"] += os.pathsep + '~/.local/bin'
+
         self.started_at: float = time.monotonic()
         self.services_running = {}
 
@@ -133,6 +159,8 @@ class BaseRunner(metaclass=ABCMeta):
     async def _init(self, cmdargs) -> None:
         self.cmdargs = cmdargs
         loop = current_loop()
+        self._service_lock = asyncio.Lock()
+
         # Initialize event loop.
         executor = concurrent.futures.ThreadPoolExecutor(max_workers=2)
         loop.set_default_executor(executor)
@@ -248,9 +276,9 @@ class BaseRunner(metaclass=ABCMeta):
         except Exception:
             log.exception('Unexpected error!')
             log.warning('We are skipping the error but the container may not work as expected.')
-            return
-        if self.init_done is not None:
-            self.init_done.set()
+        finally:
+            if self.init_done is not None:
+                self.init_done.set()
 
     @abstractmethod
     async def init_with_loop(self) -> None:
@@ -503,92 +531,105 @@ class BaseRunner(metaclass=ABCMeta):
         """Start an application service daemon."""
         return None, {}
 
-    async def _start_service(self, service_info):
-        await self._service_lock.acquire()
-        try:
-            if service_info['protocol'] == 'preopen':
-                # skip subprocess spawning as we assume the user runs it manually.
-                result = {'status': 'started'}
-                return
-            if service_info['name'] in self.services_running:
-                result = {'status': 'running'}
-                return
-            if service_info['protocol'] == 'pty':
-                result = {'status': 'failed',
-                          'error': 'not implemented yet'}
-                return
-            cwd = Path.cwd()
-            cmdargs, env = None, {}
-            if service_info['name'] == 'ttyd':
-                cmdargs, env = await prepare_ttyd_service(service_info)
-            elif service_info['name'] == 'sshd':
-                cmdargs, env = await prepare_sshd_service(service_info)
-            elif service_info['name'] == 'vscode':
-                cmdargs, env = await prepare_vscode_service(service_info)
-            elif self.service_parser is not None:
-                self.service_parser.variables['ports'] = service_info['ports']
-                cmdargs, env = await self.service_parser.start_service(
-                    service_info['name'],
-                    self.child_env.keys(),
-                    service_info['options'],
-                )
-            if cmdargs is None:
-                # fall-back to legacy service routine
-                start_info = await self.start_service(service_info)
-                if start_info is None:
-                    cmdargs, env = None, {}
-                elif len(start_info) == 3:
-                    cmdargs, env, cwd = start_info
-                elif len(start_info) == 2:
-                    cmdargs, env = start_info
-            if cmdargs is None:
-                # still not found?
-                log.warning('The service {0} is not supported.',
-                            service_info['name'])
-                result = {'status': 'failed',
-                          'error': 'unsupported service'}
-                return
-            log.debug('cmdargs: {0}', cmdargs)
-            log.debug('env: {0}', env)
-            service_env = {**self.child_env, **env}
-            # avoid conflicts with Python binary used by service apps.
-            if 'LD_LIBRARY_PATH' in service_env:
-                service_env['LD_LIBRARY_PATH'] = \
-                    service_env['LD_LIBRARY_PATH'].replace('/opt/backend.ai/lib:', '')
+    async def _start_service(self, service_info, user_requested: bool = True):
+        async with self._service_lock:
             try:
-                proc = await asyncio.create_subprocess_exec(
-                    *map(str, cmdargs),
-                    env=service_env,
-                    cwd=cwd,
-                )
-                self.services_running[service_info['name']] = proc
-                asyncio.create_task(self._wait_service_proc(service_info['name'], proc))
-                self._service_lock.release()
-                await wait_local_port_open(service_info['port'])
-                log.info("Service {} has started (pid: {}, port: {})",
-                         service_info['name'], proc.pid, service_info['port'])
-                result = {'status': 'started'}
-            except asyncio.CancelledError:
-                # This may happen if the service process gets started but it fails to
-                # open the port and then terminates (with an error).
-                result = {'status': 'failed',
-                          'error': f"the process did not start properly: {cmdargs[0]}"}
-            except PermissionError:
-                result = {'status': 'failed',
-                          'error': f"the target file is not executable: {cmdargs[0]}"}
-            except FileNotFoundError:
-                result = {'status': 'failed',
-                          'error': f"the executable file is not found: {cmdargs[0]}"}
-        except Exception as e:
-            log.exception('start_service: unexpected error')
-            result = {'status': 'failed', 'error': repr(e)}
-        finally:
-            await self.outsock.send_multipart([
-                b'service-result',
-                json.dumps(result).encode('utf8'),
-            ])
-            if self._service_lock.locked():
-                self._service_lock.release()
+                if service_info['protocol'] == 'preopen':
+                    # skip subprocess spawning as we assume the user runs it manually.
+                    result = {'status': 'started'}
+                    return
+                if service_info['name'] in self.services_running:
+                    result = {'status': 'running'}
+                    return
+                if service_info['protocol'] == 'pty':
+                    result = {'status': 'failed',
+                            'error': 'not implemented yet'}
+                    return
+                cwd = Path.cwd()
+                cmdargs: Optional[Sequence[Union[str, os.PathLike]]]
+                env: Mapping[str, str]
+                cmdargs, env = None, {}
+                if service_info['name'] == 'ttyd':
+                    cmdargs, env = await prepare_ttyd_service(service_info)
+                elif service_info['name'] == 'sshd':
+                    cmdargs, env = await prepare_sshd_service(service_info)
+                elif service_info['name'] == 'vscode':
+                    cmdargs, env = await prepare_vscode_service(service_info)
+                elif self.service_parser is not None:
+                    self.service_parser.variables['ports'] = service_info['ports']
+                    cmdargs, env = await self.service_parser.start_service(
+                        service_info['name'],
+                        self.child_env.keys(),
+                        service_info['options'],
+                    )
+                if cmdargs is None:
+                    # fall-back to legacy service routine
+                    start_info = await self.start_service(service_info)
+                    if start_info is None:
+                        cmdargs, env = None, {}
+                    elif len(start_info) == 3:
+                        cmdargs, env, cwd = start_info
+                    elif len(start_info) == 2:
+                        cmdargs, env = start_info
+                if cmdargs is None:
+                    # still not found?
+                    log.warning('The service {0} is not supported.',
+                                service_info['name'])
+                    result = {
+                        'status': 'failed',
+                        'error': 'unsupported service',
+                    }
+                    return
+                log.debug('cmdargs: {0}', cmdargs)
+                log.debug('env: {0}', env)
+                service_env = {**self.child_env, **env}
+                # avoid conflicts with Python binary used by service apps.
+                if 'LD_LIBRARY_PATH' in service_env:
+                    service_env['LD_LIBRARY_PATH'] = \
+                        service_env['LD_LIBRARY_PATH'].replace('/opt/backend.ai/lib:', '')
+                try:
+                    proc = await asyncio.create_subprocess_exec(
+                        *map(str, cmdargs),
+                        env=service_env,
+                        cwd=cwd,
+                    )
+                    self.services_running[service_info['name']] = proc
+                    asyncio.create_task(self._wait_service_proc(service_info['name'], proc))
+                    with timeout(5.0):
+                        await wait_local_port_open(service_info['port'])
+                    log.info("Service {} has started (pid: {}, port: {})",
+                             service_info['name'], proc.pid, service_info['port'])
+                    result = {'status': "started"}
+                except asyncio.CancelledError:
+                    # This may happen if the service process gets started but it fails to
+                    # open the port and then terminates (with an error).
+                    result = {'status': "failed",
+                              'error': f"the process did not start properly: {cmdargs[0]}"}
+                except asyncio.TimeoutError:
+                    # Takes too much time to open a local port.
+                    if service_info['name'] in self.services_running:
+                        await terminate_and_wait(proc, timeout=10.0)
+                        self.services_running.pop(service_info['name'], None)
+                    result = {'status': "failed",
+                              'error': f"opening the service port timed out: {service_info['name']}"}
+                except PermissionError:
+                    result = {'status': "failed",
+                              'error': f"the target file is not executable: {cmdargs[0]}"}
+                except FileNotFoundError:
+                    result = {'status': "failed",
+                              'error': f"the executable file is not found: {cmdargs[0]}"}
+            except Exception as e:
+                log.exception('start_service: unexpected error')
+                result = {
+                    'status': 'failed',
+                    'error': repr(e),
+                }
+            finally:
+                if user_requested:
+                    await self.outsock.send_multipart([
+                        b'service-result',
+                        json.dumps(result).encode('utf8'),
+                    ])
 
     async def _wait_service_proc(
         self,
@@ -603,10 +644,11 @@ class BaseRunner(metaclass=ABCMeta):
         """A thin wrapper for an external command."""
         loop = current_loop()
         if Path('/home/work/.logs').is_dir():
-            kernel_id = Path('/home/config/kernel_id.txt').read_text().strip()
+            kernel_id = os.environ['BACKENDAI_KERNEL_ID']
+            kernel_id_hex = uuid.UUID(kernel_id).hex
             log_path = Path(
                 '/home/work/.logs/task/'
-                f'{kernel_id[:2]}/{kernel_id[2:4]}/{kernel_id[4:]}.log'
+                f'{kernel_id_hex[:2]}/{kernel_id_hex[2:4]}/{kernel_id_hex[4:]}.log'
             )
             log_path.parent.mkdir(parents=True, exist_ok=True)
         else:
@@ -651,6 +693,15 @@ class BaseRunner(metaclass=ABCMeta):
     async def shutdown(self):
         pass
 
+    async def _shutdown_service(self, service_name: str):
+        try:
+            async with self._service_lock:
+                if service_name in self.services_running:
+                    await terminate_and_wait(self.services_running[service_name])
+                    self.services_running.pop(service_name, None)
+        except Exception:
+            log.exception('unexpected error (shutdown_service)')
+
     async def handle_user_input(self, reader, writer):
         try:
             if self.user_input_queue is None:
@@ -721,6 +772,26 @@ class BaseRunner(metaclass=ABCMeta):
         if user_bootstrap_path.is_file():
             await self._bootstrap(user_bootstrap_path)
 
+        log.debug('starting intrinsic services: sshd, ttyd ...')
+        intrinsic_spawn_coros = []
+        intrinsic_spawn_coros.append(self._start_service({
+            'name': 'sshd',
+            'port': 2200,
+            'protocol': 'tcp',
+        }, user_requested=False))
+        intrinsic_spawn_coros.append(self._start_service({
+            'name': 'ttyd',
+            'port': 7681,
+            'protocol': 'http',
+        }, user_requested=False))
+        results = await asyncio.gather(*intrinsic_spawn_coros, return_exceptions=True)
+        for result in results:
+            if isinstance(result, Exception):
+                log.exception(
+                    'error during starting intrinsic services',
+                    exc_info=result,
+                )
+
         log.debug('start serving...')
         while True:
             try:
@@ -751,6 +822,9 @@ class BaseRunner(metaclass=ABCMeta):
                 elif op_type == 'start-service':  # activate a service port
                     data = json.loads(text)
                     asyncio.create_task(self._start_service(data))
+                elif op_type == 'shutdown-service':  # shutdown the service by its name
+                    data = json.loads(text)
+                    await self._shutdown_service(data)
                 elif op_type == 'get-apps':
                     await self._get_apps(text)
             except asyncio.CancelledError:
